import logging
import sys

import typer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
logger.addHandler(handler)

cli = typer.Typer()



import json
from collections.abc import Generator
from glob import glob
from pathlib import Path
from time import time

import requests
import socketio
import typer
from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from fernet_encrypt import logger

KEY_PATH = Path(__file__).parent / "keys"


def key_getter() -> Generator[bytes, None, None]:
    keyfiles = sorted(glob(str(KEY_PATH / "*.key")), reverse=True)

    if len(keyfiles) == 0:
        raise Exception("No keyfiles found. Create a key first.")

    for keyfile in keyfiles:
        with open(keyfile, "rb") as f:
            yield f.read()


def key_setter(key: bytes):
    keyfile = KEY_PATH / f"{int(time())}.key"

    with open(keyfile, "wb") as f:
        f.write(key)

    logger.debug(f"Created keyfile: {keyfile}")


class API:
    REMOTE_API = "https://fernet-encrypt.tysonholub.com"

    def __init__(self):
        self.token = self._set_token()
        self.session = Session()
        self.session.headers = {"Authorization": f"Bearer {self.token['access_token'] if self.token else ''}"}
        retries = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[500, 502, 503, 504],
        )
        adapter = HTTPAdapter(max_retries=retries)
        self.session.mount(self.REMOTE_API, adapter)

        self.sio = socketio.Client()
        self.sio.on("logged_in", self._handle_logged_in)
        self.sio.on("login", self._handle_login)

    def _set_token(self) -> dict | None:
        try:
            with open(Path.home() / ".fernet-encrypt") as f:
                token = json.load(f)
        except FileNotFoundError:
            token = None

        return token

    def _prepare_request(self, method, url, **kwargs) -> PreparedRequest:
        req = requests.Request(method, url, **kwargs)
        prepared_req = self.session.prepare_request(req)

        return prepared_req

    def _make_request(self, prepared_req: PreparedRequest) -> Response:
        if not self.token:
            self.login()
            self.token = self._set_token()
            self._refresh_token(prepared_req=prepared_req)

        response = self.session.send(prepared_req, timeout=5)

        if response.status_code == 401:
            self._refresh_token(prepared_req=prepared_req)
            response = self.session.send(prepared_req, timeout=5)

        response.raise_for_status()

        return response

    def _refresh_token(self, prepared_req: PreparedRequest | None = None):
        if not self.token:
            raise Exception("No token to refresh. Have you logged in?")

        refresh_token = self.token["refresh_token"]
        response = self.session.post(
            f"{self.REMOTE_API}/cli/refresh_token",
            json={"token": refresh_token},
            headers={"Content-Type": "application/json"},
            timeout=5,
        )

        response.raise_for_status()

        self.token = response.json()
        self.token.update({
            "refresh_token": refresh_token,
        })

        self.session.headers.update({"Authorization": f"Bearer {self.token['access_token']}"})
        if prepared_req:
            prepared_req.headers.update({"Authorization": f"Bearer {self.token['access_token']}"})

        with open(Path.home() / ".fernet-encrypt", "w+") as f:
            json.dump(self.token, f)

    def _handle_login(self, data):
        typer.echo(f"Login in here: {data['login_url']}")

    def _handle_logged_in(self, data):
        with open(Path.home() / ".fernet-encrypt", "w+") as f:
            json.dump(data["token"], f)

        self.sio.disconnect()

    def login(self):
        self.sio.connect(self.REMOTE_API)
        self.sio.emit("login")
        self.sio.wait()

    def encrypt(self, input: bytes, filename: str) -> bytes:
        prepared_req = self._prepare_request(
            method="POST",
            url=f"{self.REMOTE_API}/api/encrypt",
            json={
                "file": input.decode("utf-8"),
                "filename": filename,
            },
        )
        return self._make_request(prepared_req).content

    def decrypt(self, input: bytes, filename: str) -> bytes:
        prepared_req = self._prepare_request(
            method="POST",
            url=f"{self.REMOTE_API}/api/decrypt",
            json={
                "encrypted_file": input.decode("utf-8"),
                "filename": filename,
            },
        )

        return self._make_request(prepared_req).content

    def get_keys(self) -> dict:
        prepared_req = self._prepare_request(
            method="GET",
            url=f"{self.REMOTE_API}/api/keys",
        )

        return self._make_request(prepared_req).json()


api = API()



from collections.abc import Callable, Generator

from cryptography.fernet import Fernet

from fernet_encrypt.utils import api, key_getter, key_setter


def create_key(key_setter: Callable = key_setter):
    key = Fernet.generate_key()
    key_setter(key)


def encrypt(input: bytes, key_getter: Callable[[], Generator[bytes, None, None]] = key_getter) -> bytes:
    key = Fernet(next(key_getter()))
    return key.encrypt(input)


def decrypt(input: bytes, key_getter: Callable[[], Generator[bytes, None, None]] = key_getter) -> bytes:
    key_count = 0
    for key_str in key_getter():
        key = Fernet(key_str)
        key_count += 1

        try:
            decrypted_data = key.decrypt(input)
        except Exception:  # nosec: CWE-703
            continue

        return decrypted_data

    raise Exception(f"Unable to decrypt input with {key_count} existing keys.")


def remote_login():
    api.login()


def get_remote_keys() -> dict:
    return api.get_keys()


def decrypt_remote(input: bytes, filename: str) -> bytes:
    return api.decrypt(input, filename)


def encrypt_remote(input: bytes, filename: str) -> bytes:
    return api.encrypt(input, filename)



import json
from pathlib import Path
from typing import Annotated

import typer

from fernet_encrypt import cli
from fernet_encrypt.commands import (
    create_key,
    decrypt,
    decrypt_remote,
    encrypt,
    encrypt_remote,
    get_remote_keys,
    remote_login,
)
from fernet_encrypt.utils import KEY_PATH, api, logger


@cli.command(name="login", help=f"Sign up or Login to {api.REMOTE_API}")
def login():
    remote_login()


@cli.command(name="get-keys", help="Sync remote keys to fernet-encrypt local key storage")
def get_keys(
    stdout: Annotated[bool, typer.Option("--stdout", help="Output fetched keys as JSON to stdout")] = False,
):
    keys = get_remote_keys()

    if stdout:
        typer.echo(json.dumps(keys, indent=2))

    for key in keys:
        keyfile = KEY_PATH / f"{key['timestamp']}.key"
        with open(keyfile, "wb") as f:
            f.write(key["key"].encode("utf-8"))

        logger.debug(f"Created keyfile: {keyfile}")


@cli.command(name="create-local-key", help="Create a new local storage Fernet key")
def create_local_key():
    create_key()


@cli.command(name="encrypt-file", help="Encrypt a local file using local key storage or remote API keys")
def encrypt_file(
    input_file: Path = typer.Argument(default=..., exists=True, help="The input file to encrypt"),
    remote: Annotated[bool, typer.Option("--remote", help="Encrypt using the remote API")] = False,
):
    with open(input_file, "rb") as f:
        input_data = f.read()

    if remote:
        result = encrypt_remote(input_data, input_file.name)
    else:
        result = encrypt(input_data)

    typer.echo(result, nl=False)


@cli.command(name="decrypt-file", help="Decrypt a local file using local key storage or remote API keys")
def decrypt_file(
    input_file: Path = typer.Argument(default=..., exists=True, help="The input file to decrypt"),
    remote: Annotated[bool, typer.Option("--remote", help="Decrypt using the remote API")] = False,
):
    with open(input_file, "rb") as f:
        input_data = f.read()

    if remote:
        result = decrypt_remote(input_data, input_file.name)
    else:
        result = decrypt(input_data)

    typer.echo(result, nl=False)
